<?php

namespace Drupal\ldap_sso\StackMiddleware;

use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\PageCache\RequestPolicyInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\ldap_servers\Logger\LdapDetailLog;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\HttpKernelInterface;

/**
 * Executes the page caching before the main kernel takes over the request.
 */
class SeamlessAuthentication implements HttpKernelInterface {

  /**
   * The wrapped HTTP kernel.
   *
   * @var \Symfony\Component\HttpKernel\HttpKernelInterface
   */
  protected $httpKernel;

  /**
   * A policy rule determining the cacheability of a request.
   *
   * @var \Drupal\Core\PageCache\RequestPolicyInterface
   */
  protected $requestPolicy;

  /**
   * LDAP SSO config.
   *
   * @var \Drupal\Core\Config\Config|\Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Front page.
   *
   * @var array|mixed|null
   */
  protected $frontpage;

  /**
   * Detail log.
   *
   * @var \Drupal\ldap_servers\Logger\LdapDetailLog
   */
  protected $detailLog;

  /**
   * The current user account.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $account;

  /**
   * Default paths to exclude.
   *
   * @var string[]
   */
  protected const DEFAULT_EXCLUDE_PATHS = [
    '/admin/config/search/clean-urls/check',
    '/user/login/sso',
    '/user/login',
    '/user/logout',
    '/user',
  ];

  /**
   * SeamlessAuthentication constructor.
   *
   * @param \Symfony\Component\HttpKernel\HttpKernelInterface $http_kernel
   *   Kernel.
   * @param \Drupal\Core\PageCache\RequestPolicyInterface $request_policy
   *   Request policy.
   * @param \Drupal\Core\Config\ConfigFactory $configFactory
   *   Config factory.
   * @param \Drupal\ldap_servers\Logger\LdapDetailLog $detailLog
   *   Detail log.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   Account.
   */
  public function __construct(
    HttpKernelInterface $http_kernel,
    RequestPolicyInterface $request_policy,
    ConfigFactory $configFactory,
    LdapDetailLog $detailLog,
    AccountInterface $account
  ) {
    $this->httpKernel = $http_kernel;
    $this->requestPolicy = $request_policy;
    $this->config = $configFactory->get('ldap_sso.settings');
    $this->frontpage = $configFactory->get('system.site')->get('frontpage');
    $this->detailLog = $detailLog;
    $this->account = $account;
  }

  /**
   * {@inheritdoc}
   */
  public function handle(Request $request, $type = self::MASTER_REQUEST, $catch = TRUE): Response {

    if (
      $type === static::MASTER_REQUEST &&
      $this->requestPolicy->check($request) === RequestPolicyInterface::ALLOW &&
      $this->lookup($request, $type, $catch)
    ) {
      $server_variables = $request->server->all();
      $server_variables['REQUEST_URI'] = $request->getBasePath() . '/user/login/sso';
      $request = $request->duplicate(
        ['destination' => $request->getBasePath() . $request->getPathInfo()],
        NULL,
        NULL,
        NULL,
        NULL,
        $server_variables
      );
    }

    return $this->httpKernel->handle($request, $type, $catch);
  }

  /**
   * Retrieves a response from the cache or fetches it from the backend.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   A request object.
   *
   * @returns \Symfony\Component\HttpFoundation\Response $response
   *   A response object.
   */
  protected function lookup(Request $request): bool {
    if (PHP_SAPI === 'cli') {
      $this->detailLog->log('CLI, no SSO.', [], 'ldap_sso');
      return FALSE;
    }

    if (!$this->config->get('seamlessLogin')) {
      $this->detailLog->log('Automated SSO not active.', [], 'ldap_sso');
      return FALSE;
    }

    if ($this->checkExcludePath($request)) {
      $this->detailLog->log(
        'Excluded URL @path',
        ['@path' => $request->getUri()],
        'ldap_sso'
      );
      return FALSE;
    }

    if ($request->cookies->get('sso_stop')) {
      $this->detailLog->log('Anonymous user with cookie to not continue SSO login', [], 'ldap_sso');
      return FALSE;
    }

    $this->detailLog->log('Transferring to login controller', [], 'ldap_sso');
    return TRUE;
  }

  /**
   * Check to exclude paths from SSO.
   *
   * @return bool
   *   Path excluded or not.
   */
  protected function checkExcludePath(Request $request): bool {
    if (in_array($request->getPathInfo(), self::DEFAULT_EXCLUDE_PATHS, TRUE)) {
      return TRUE;
    }

    if (is_array($this->config->get('ssoExcludedHosts'))) {
      $host = $_SERVER['SERVER_NAME'];
      foreach ($this->config->get('ssoExcludedHosts') as $host_to_check) {
        if ($host_to_check === $host) {
          return TRUE;
        }
      }
    }

    foreach ($this->config->get('ssoExcludedPaths') as $path_to_exclude) {
      if (
        mb_strtolower($path) === mb_strtolower($path_to_exclude) ||
        ($path_to_exclude === '<front>' && mb_strtolower($this->frontpage) === mb_strtolower($path))
      ) {
        return TRUE;
      }
    }

    return FALSE;
  }

}
